#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <xc.h>
#include <sys/attribs.h>
#include <sys/kmem.h>
#include "FT8xx.h"
#include "Library.h"
#include "function_prototypes.h"
#include "CAN_vars.h"
#include "global.h"


void initCAN(void){
    
    /* Turn CAN off to Reset it */
    C2CONbits.REQOP = 4;    /* Place the CAN module in Configuration mode. */
    while(C2CONbits.OPMOD != 4);
    
    C2CONCLR = 0x00008000;        /* Switch the CAN module off by clearing the ON bit */
    while(C2CONbits.CANBUSY == 1);
    
    /* Switch the CAN module back ON and switch it to Configuration mode. Wait till the switch is complete */
    C2CONbits.ON = 1;
    C2CONbits.REQOP = 4;
    while(C2CONbits.OPMOD != 4);
    
    /* Configure the CAN Module Clock */
    set_Baud();
    
    /* Set CAN2_RX pin as input */
    AD1PCFG = 0x00004100;       /* Disable Analog Mode */
    TRISBbits.TRISB14 = 1;
    
    /* Put it back in normal mode */
    C2CONbits.REQOP = 0;
    while(C2CONbits.OPMOD != 0);

    /* Configure filters and mask */
    initFilterRight();
    initFilterLeft();
    
}

void set_Baud(void){
/* Example 34-17: Configuring the CAN Module to Obtain a Specific Bit Rate */

/* This code example shows how to configure the CAN module to obtain a */
/* specific bit rate implementing the configuration shown in Example 34-16. */
/* Fsys = System Clock Frequency = 80MHz; */
/* Fbaud = CAN bit rate = 250000; */
/* N = Time Quanta (Tq) per bit = 16; */
/* Prop Segment = 1 Tq */
/* Phase Seg 1 = 7 Tq */
/* Phase Seg 2 = 7 Tq */
/* Sync Jump Width = 2 Tq */

/* Ensure the CAN module is in configuration mode.*/

    C2CONbits.REQOP = 4;
    while(C2CONbits.OPMOD != 4);

    C2CFGbits.SEG2PHTS = 1;     /* Phase seg 2 is freely programmable */
    C2CFGbits.SEG2PH = 6;       /* Phase seg 2 is 7 Tq.*/ //cause problems with EFL
    C2CFGbits.SEG1PH = 6;       /* Phase seg 1 is 7 Tq.*/ //cause problems with EFL
    C2CFGbits.PRSEG = 0;        /* Propagation seg 2 is 1 Tq. */
    C2CFGbits.SAM = 0;          /* Sample bit 1 time. */
    C2CFGbits.SJW = 2;          /* Sync jump width is 2 Tq */
    C2CFGbits.BRP = 9;         /* BRP is (Fsys/(2*Ftq))-1 = (80e6/(2*4e6))-1 = 9 */
}

void initFIFO(){
/* Allocate a total of 16 words */
    unsigned int CanFifoMessageBuffers[36]; //changes to 36 from 32

/* Request CAN to switch to configuration mode and wait until it has switched */
    C2CONbits.REQOP = 4;
    while(C2CONbits.OPMOD != 4);

/* Initialize C2FIFOBA register with physical address of CAN message Buffer */
    C2FIFOBA = KVA_TO_PA(CanFifoMessageBuffers);

/* Configure FIFO0. This will be a receive FIFO for 4 full messages */
    C2FIFOCON0bits.FSIZE = 3;   /* size is 4 message buffers (MB0 to MB3) */
    C2FIFOCON0bits.TXEN = 0;    /* Clear the TXEN bit */
    C2FIFOCON0bits.DONLY = 0;   /* full message */
    
/* Configure FIFO1. This will be a receive FIFO for 4 full messages */
    C2FIFOCON1bits.FSIZE = 3;   /* size is 4 message buffers (MB0 to MB3) */
    C2FIFOCON1bits.TXEN = 0;    /* Clear the TXEN bit */
    C2FIFOCON1bits.DONLY = 0;   /* full message */

    
/* Configure FIFO2. This will be a receive FIFO for 4 full messages */
    C2FIFOCON2bits.FSIZE = 3;   /* size is 4 message buffers (MB0 to MB3) */
    C2FIFOCON2bits.TXEN = 0;    /* Clear the TXEN bit */
    C2FIFOCON2bits.DONLY = 0;   /* full message */

    
/* Configure FIFO3. This will be a receive FIFO for 4 full messages */
    C2FIFOCON3bits.FSIZE = 3;   /* size is 4 message buffers (MB0 to MB3) */
    C2FIFOCON3bits.TXEN = 0;    /* Clear the TXEN bit */
    C2FIFOCON3bits.DONLY = 0;   /* full message */


/* The CAN module can now be placed into normal mode if no further
 *  configuration is required. */
    C2CONbits.REQOP = 0;
    while(C2CONbits.OPMOD != 0);
}

struct CAN_out Read_FIFO(void){
    struct CAN_out CANOUT;
    struct CAN_out0_R CANOUT_0_R;
    struct CAN_out1_R CANOUT_1_R;
    struct CAN_out0_L CANOUT_0_L;
    struct CAN_out1_L CANOUT_1_L;
/* From Example 34-4: Reading Received Messages from the FIFO */
/* FIFO0 size is 8 messages and each message is 4 words long. */
    unsigned int * currentMessageBuffer; /* Points to message buffer to be read */
    while(C2CONbits.ON == 1){
    /* MESSAGE 1 */
        /* Keep reading until the FIFO0 is empty. */
        while(C2FIFOINT0bits.RXNEMPTYIF == 1){
            /* Get the address of the message buffer to read from the C2FIFOUA0
             * register. Convert this physical address to virtual address. */
                currentMessageBuffer = PA_TO_KVA1(C2FIFOUA0);
                CANOUT_0_R = ProcessMessage0(currentMessageBuffer);
                CANOUT.rpm_R = CANOUT_0_R.rpm_R;
                CANOUT.motor_temp_R = CANOUT_0_R.motor_current_R;
                CANOUT.bat_voltage_R = CANOUT_0_R.bat_voltage_R;
                CANOUT.error_lsb_R = CANOUT_0_R.error_lsb_R;
                CANOUT.error_msb_R = CANOUT_0_R.error_msb_R;
            /* Set the UINC bit to tell the CAN module that
             * a message has been read. */
                C2FIFOCON0bits.UINC = 1;   
        }
        
    /* MESSAGE 2 */
    /* Keep reading until the FIFO1 is empty. */
        while(C2FIFOINT1bits.RXNEMPTYIF == 1){
            /* Get the address of the message buffer to read from the C2FIFOUA1
             * register. Convert this physical address to virtual address. */
                currentMessageBuffer = PA_TO_KVA1(C2FIFOUA1);
                CANOUT_1_R = ProcessMessage1(currentMessageBuffer);
                CANOUT.contr_temp_R = CANOUT_1_R.contr_temp_R;
                CANOUT.contr_status_R = CANOUT_1_R.contr_status_R;
                CANOUT.switch_status_R = CANOUT_1_R.switch_status_R;
                CANOUT.motor_temp_R = CANOUT_1_R.motor_temp_R;
                CANOUT.throttle_R = CANOUT_1_R.throttle_R;
            /* Set the UINC bit to tell the CAN module that
             * a message has been read. */
            C2FIFOCON1bits.UINC = 1;
        }
        
    /* MESSAGE 3 */
    /* Keep reading until the FIFO2 is empty. */
        while(C2FIFOINT2bits.RXNEMPTYIF == 1){
            /* Get the address of the message buffer to read from the C2FIFOUA2
             * register. Convert this physical address to virtual address. */
                currentMessageBuffer = PA_TO_KVA1(C2FIFOUA2);
                CANOUT_0_L = ProcessMessage2(currentMessageBuffer);
                CANOUT.rpm_L = CANOUT_0_L.rpm_L;
                CANOUT.motor_temp_L = CANOUT_0_L.motor_current_L;
                CANOUT.bat_voltage_L = CANOUT_0_L.bat_voltage_L;
                CANOUT.error_lsb_L = CANOUT_0_L.error_lsb_L;
                CANOUT.error_msb_L = CANOUT_0_L.error_msb_L;
            /* Set the UINC bit to tell the CAN module that
             * a message has been read. */
            C2FIFOCON2bits.UINC = 1;
        }
        
    /* MESSAGE 4 */
    /* Keep reading until the FIFO3 is empty. */
        while(C2FIFOINT3bits.RXNEMPTYIF == 1){
            /* Get the address of the message buffer to read from the C2FIFOUA3
             * register. Convert this physical address to virtual address. */
                currentMessageBuffer = PA_TO_KVA1(C2FIFOUA3);
                CANOUT_1_L = ProcessMessage3(currentMessageBuffer);
                CANOUT.contr_temp_L = CANOUT_1_L.contr_temp_L;
                CANOUT.contr_status_L = CANOUT_1_L.contr_status_L;
                CANOUT.switch_status_L = CANOUT_1_L.switch_status_L;
                CANOUT.motor_temp_L = CANOUT_1_L.motor_temp_L;
                CANOUT.throttle_L = CANOUT_1_L.throttle_L;            

            /* Set the UINC bit to tell the CAN module that
             * a message has been read. */
            C2FIFOCON3bits.UINC = 1;
        }
    }
return CANOUT;
}

struct CAN_out0_R ProcessMessage0(int currentMessageBuffer){
    CANRxMessageBuffer *buffer;
/* When a message have been received and read, the individual fields of */
/* the received message can be queried as such. */
    buffer = (CANRxMessageBuffer *) (PA_TO_KVA1(C2FIFOUA0));
    
    if(buffer->CMSGEID.DLC == 8){
    /* If the length of the received message is 8 then process the message. */
        unsigned int * bufferToRead;
    /* Check if there is a message available to read. */
        if(C2FIFOINT0bits.RXNEMPTYIF == 1){
        /* Get the address of the buffer to read */
            bufferToRead = PA_TO_KVA1(C2FIFOUA0);
            
            EID = buffer->CMSGEID.EID;
            int D1 = buffer->CMSGDATA0.Byte0;
            int D2 = buffer->CMSGDATA0.Byte1;
            float D3 = buffer->CMSGDATA0.Byte2;
            float D4 = buffer->CMSGDATA0.Byte3;

            int D5 = buffer->CMSGDATA1.Byte0;
            int D6 = buffer->CMSGDATA1.Byte1;
            int D7 = buffer->CMSGDATA1.Byte2;
            int D8 = buffer->CMSGDATA1.Byte3;

            rpm_R = (D2*256)+D1;
            motor_current_R = ((D4*256)+D3)/10;
            bat_voltage_R = ((D6*256)+D5)/10;
            error_lsb_R = D7;
            error_msb_R = D8;
                        
            // Store Full message 1 //
            msg_array[0] = buffer->messageWord[0];
            msg_array[1] = buffer->messageWord[1];
            msg_array[2] = buffer->messageWord[2];
            msg_array[3] = buffer->messageWord[3];

            // define CAN outputs
            struct CAN_out0_R CAN_OUT0_R = {rpm_R, motor_current_R, bat_voltage_R, error_lsb_R, error_msb_R};

        /* Update the message buffer pointer. */
            C2FIFOCON0bits.UINC = 1;
           
        /* Turn CAN off */
            C2CONbits.REQOP = 4;    /* Place the CAN module in Configuration mode. */
            while(C2CONbits.OPMOD != 4);
            C2CONbits.ON = 0;
            while(C2CONbits.ON != 0);
        return CAN_OUT0_R;
        }
    }
}
struct CAN_out1_R ProcessMessage1(int currentMessageBuffer){
    CANRxMessageBuffer *buffer;  
/* When a message have been received and read, the individual fields of */
/* the received message can be queried as such. */
    buffer = (CANRxMessageBuffer *) (PA_TO_KVA1(C2FIFOUA1));
    
    if(buffer->CMSGEID.DLC == 8){
    /* If the length of the received message is 8 then process the message. */
        unsigned int * bufferToRead;
    /* Check if there is a message available to read. */
        if(C2FIFOINT1bits.RXNEMPTYIF == 1){
        /* Get the address of the buffer to read */
            bufferToRead = PA_TO_KVA1(C2FIFOUA1);
            
            EID = buffer->CMSGEID.EID;
            float D1 = buffer->CMSGDATA0.Byte0;
            int D2 = buffer->CMSGDATA0.Byte1;
            int D3 = buffer->CMSGDATA0.Byte2;
            int D4 = buffer->CMSGDATA0.Byte3;

            int D5 = buffer->CMSGDATA1.Byte0;
            int D6 = buffer->CMSGDATA1.Byte1;
            int D7 = buffer->CMSGDATA1.Byte2;
            int D8 = buffer->CMSGDATA1.Byte3;
            
            throttle_R = (D1/255)*100;
            contr_temp_R = D2 - 40;
            motor_temp_R = D3 - 30;
            contr_status_R = D5;
            switch_status_R = D6;
                        
            // Store Full message 2 //
            msg_array[4] = buffer->messageWord[0];
            msg_array[5] = buffer->messageWord[1];
            msg_array[6] = buffer->messageWord[2];
            msg_array[7] = buffer->messageWord[3];

            // define CAN outputs
            struct CAN_out1_R CAN_OUT1_R = {throttle_R, contr_temp_R, motor_temp_R, contr_status_R, switch_status_R};
        /* Update the message buffer pointer. */
            C2FIFOCON1bits.UINC = 1;
            
        /* Turn CAN off */
            C2CONbits.REQOP = 4;    /* Place the CAN module in Configuration mode. */
            while(C2CONbits.OPMOD != 4);
            C2CONbits.ON = 0;
            while(C2CONbits.ON != 0);
            return CAN_OUT1_R;
        }
    }
}
struct CAN_out0_L ProcessMessage2(int currentMessageBuffer){
    CANRxMessageBuffer *buffer;

/* When a message have been received and read, the individual fields of */
/* the received message can be queried as such. */
    buffer = (CANRxMessageBuffer *) (PA_TO_KVA1(C2FIFOUA2));
    
    if(buffer->CMSGEID.DLC == 8){
    /* If the length of the received message is 8 then process the message. */
        unsigned int * bufferToRead;
    /* Check if there is a message available to read. */
        if(C2FIFOINT2bits.RXNEMPTYIF == 1){
        /* Get the address of the buffer to read */
            bufferToRead = PA_TO_KVA1(C2FIFOUA2);
            
            EID = buffer->CMSGEID.EID;
            int D1 = buffer->CMSGDATA0.Byte0;
            int D2 = buffer->CMSGDATA0.Byte1;
            float D3 = buffer->CMSGDATA0.Byte2;
            float D4 = buffer->CMSGDATA0.Byte3;

            int D5 = buffer->CMSGDATA1.Byte0;
            int D6 = buffer->CMSGDATA1.Byte1;
            int D7 = buffer->CMSGDATA1.Byte2;
            int D8 = buffer->CMSGDATA1.Byte3;

            rpm_L = (D2*256)+D1;
            motor_current_L = ((D4*256)+D3)/10;
            bat_voltage_L = ((D6*256)+D5)/10;
            error_lsb_L = D7;
            error_msb_L = D8;
            // define CAN outputs
            struct CAN_out0_L CAN_OUT0_L = {rpm_L, motor_current_L, bat_voltage_L, error_lsb_L, error_msb_L};
        /* Update the message buffer pointer. */
            C2FIFOCON2bits.UINC = 1;
            
        /* Turn CAN off */
            C2CONbits.REQOP = 4;    /* Place the CAN module in Configuration mode. */
            while(C2CONbits.OPMOD != 4);
            C2CONbits.ON = 0;
            while(C2CONbits.ON != 0);
            return CAN_OUT0_L;
        }
    }
}
struct CAN_out1_L ProcessMessage3(int currentMessageBuffer){
    CANRxMessageBuffer *buffer;
    
/* When a message have been received and read, the individual fields of */
/* the received message can be queried as such. */
    buffer = (CANRxMessageBuffer *) (PA_TO_KVA1(C2FIFOUA3));
    
    if(buffer->CMSGEID.DLC == 8){
    /* If the length of the received message is 8 then process the message. */
        unsigned int * bufferToRead;
    /* Check if there is a message available to read. */
        if(C2FIFOINT3bits.RXNEMPTYIF == 1){
        /* Get the address of the buffer to read */
            bufferToRead = PA_TO_KVA1(C2FIFOUA3);
            
            EID = buffer->CMSGEID.EID;
            float D1 = buffer->CMSGDATA0.Byte0;
            int D2 = buffer->CMSGDATA0.Byte1;
            int D3 = buffer->CMSGDATA0.Byte2;
            int D4 = buffer->CMSGDATA0.Byte3;

            int D5 = buffer->CMSGDATA1.Byte0;
            int D6 = buffer->CMSGDATA1.Byte1;
            int D7 = buffer->CMSGDATA1.Byte2;
            int D8 = buffer->CMSGDATA1.Byte3;
            
            throttle_L = (D1/255)*100;
            contr_temp_L = D2 - 40;
            motor_temp_L = D3 - 30;
            contr_status_L = D5;
            switch_status_L = D6;
            // define CAN outputs
            struct CAN_out1_L CAN_OUT1_L = {throttle_L, contr_temp_L, motor_temp_L, contr_status_L, switch_status_L};
        /* Update the message buffer pointer. */
            C2FIFOCON3bits.UINC = 1;
            
        /* Turn CAN off */
            C2CONbits.REQOP = 4;    /* Place the CAN module in Configuration mode. */
            while(C2CONbits.OPMOD != 4);
            C2CONbits.ON = 0;
            while(C2CONbits.ON != 0);
            return CAN_OUT1_L;
        }
    }
}
void initFilterRight(void){
    /* Filter 0 is set up to accept messages with Kelly controller EID. 
     * Accepted messages will be stored in FIFO0. Mask 0 is
     * used to implement the filter address range. */
    
    /* Check that it's in in configuration mode */
    C2CONbits.REQOP = 4;
    while(C2CONbits.OPMOD != 4);    
    
    /* Kelly Message 1 Filter */
    
    C2FLTCON0bits.FSEL0 = 0x0000;           /* Store messages in FIFO0 */
    C2FLTCON0bits.MSEL0 = 0x00;             /* Use Mask 0 */
    
    C2RXF0bits.SID = 0b01100111100;         /* Kelly SID */
    C2RXF0bits.EID = 0b010001111000000110;  /* Kelly message 1 EID*/

    
    
    C2RXF0bits.EXID = 1;                    /* Match only messages with extended identifier addresses */

    C2RXM0bits.EID = 0b111111111111111111;  /* Allow only EID bits */
    C2RXM0bits.MIDE = 1;                    /* Match only EID message types. */
    
    C2FLTCON0bits.FLTEN0 = 1;               /* Enable the filter */
    
    /* Message 2 Filter */
    
    C2FLTCON0bits.FSEL1 = 0x0001;           /* Store messages in FIFO1 */
    C2FLTCON0bits.MSEL1 = 0x00;             /* Use Mask 0 (already configured) */
    
    C2RXF1bits.SID = 0b01100111100;         /* Kelly SID*/    
    C2RXF1bits.EID = 0b010001111100000110;  /* Kelly message 2 EID */
    
    
    C2RXF1bits.EXID = 1;                    /* Match only messages with extended identifier addresses */
    
    C2FLTCON0bits.FLTEN1 = 1;               /* Enable the filter */
    
    /* Place CAN Module in normal mode.*/
    C2CONbits.REQOP = 0;
    while(C2CONbits.OPMOD != 0);
}
void initFilterLeft(void){
    /* Filter 0 is set up to accept messages with Kelly controller EID. 
     * Accepted messages will be stored in FIFO0. Mask 0 is
     * used to implement the filter address range. */
    
    /* Check that it's in in configuration mode */
    C2CONbits.REQOP = 4;
    while(C2CONbits.OPMOD != 4);    
    
    /* Kelly Message 1 Filter */
    
    C2FLTCON0bits.FSEL2 = 0x0002;           /* Store messages in FIFO2 */
    C2FLTCON0bits.MSEL2 = 0x00;             /* Use Mask 0 */
    
    C2RXF2bits.SID = 0b01100111100;         /* Kelly SID */
    C2RXF2bits.EID = 0b010001111000000101;  /* Kelly message 1 EID */

    C2RXF2bits.EXID = 1;  /* Match only messages with extended identifier addresses */   
    
    C2FLTCON0bits.FLTEN2 = 1;               /* Enable the filter */
    
    /* Message 2 Filter */
    
    C2FLTCON0bits.FSEL3 = 0x0003;           /* Store messages in FIFO3 */
    C2FLTCON0bits.MSEL3 = 0x00;             /* Use Mask 0 (already configured) */
    
    C2RXF3bits.SID = 0b01100111100;         /* Kelly SID */    
    C2RXF3bits.EID = 0b010001111100000101;  /* Kelly message 2 EID */
    
    C2RXF3bits.EXID = 1;                    /* Match only messages with extended identifier addresses */
    
    C2FLTCON0bits.FLTEN3 = 1;               /* Enable the filter */
    
    /* Place CAN Module in normal mode.*/
    C2CONbits.REQOP = 0;
    while(C2CONbits.OPMOD != 0);
}


/* Timer 3 Interrupt Service Routine */
/* vector number for T3 is 12, Priority 6 */
void __ISR(12, IPL6AUTO) Timer3Hand(void){
    T3CONbits.ON = 0; // turn off timer3
    initCAN();
    initFIFO();
    struct CAN_out CANOUT = Read_FIFO();
    rpm = CANOUT.rpm_R;
    rpm_left = CANOUT.rpm_L;
    temp = CANOUT.motor_temp_R;
    bat_volt = CANOUT.bat_voltage_R;
    error_msb_left = CANOUT.error_msb_L;
    error_msb_right = CANOUT.error_msb_R;
    error_lsb_left = CANOUT.error_lsb_L;
    error_lsb_right = CANOUT.error_lsb_R;
//Testing Only

    if ( CANOUT.rpm_L >= 10 | CANOUT.rpm_R >= 10) {
        LATFbits.LATF1 = 1;
    }
    else {
        LATFbits.LATF1 = 0;
    }
 
    IFS0bits.T3IF = 0; // clear interrupt flag
    T3CONbits.ON = 1;
}

